(******************************************************************************)
(*                                                                            *)
(*                               PROJECT HELM                                 *)
(*                                                                            *)
(*                Claudio Sacerdoti Coen <sacerdot@cs.unibo.it>               *)
(*                                 24/01/2000                                 *)
(*                                                                            *)
(* This module implements a trival cache system (an hash-table) for cic       *)
(* objects. Uses the getter (getter.ml) and the parser (cicParser.ml)         *)
(*                                                                            *)
(******************************************************************************)

let raise e = print_endline "***" ; flush stdout ; print_endline (Printexc.to_string e) ; flush stdout ; raise e;;

(*CSC: forse i due seguenti tipi sono da unificare? *)
type cooked_obj =
   Cooked of Cic.obj
 | Frozen of Cic.obj
 | Unchecked of Cic.obj
type type_checked_obj =
   CheckedObj of Cic.obj     (* cooked obj *)
 | UncheckedObj of Cic.obj   (* uncooked obj *)
;;

exception NoFunctionProvided;;

(* CSC: da sostituire con un (...) option ref *)
let cook_obj = ref (fun obj uri -> raise NoFunctionProvided);;

exception CircularDependency of string;;
exception CouldNotUnfreeze of string;;
exception Impossible;;
exception UncookedObj;;

module HashedType =
 struct
  type t = UriManager.uri * int    (* uri, livello di cottura *)
  let equal (u1,n1) (u2,n2) = UriManager.eq u1 u2 && n1 = n2
  let hash = Hashtbl.hash
 end
;;

(* Hashtable that uses == instead of = for testing equality *)
module HashTable = Hashtbl.Make(HashedType);;

let hashtable = HashTable.create 271;;

(* n is the number of time that the object must be cooked *)
let get_obj_and_type_checking_info uri n =
 try
   HashTable.find hashtable (uri,n)
 with
  Not_found -> 
   try
    match HashTable.find hashtable (uri,0) with
        Cooked _
      | Frozen _ -> raise Impossible
      | Unchecked _ as t -> t
   with
    Not_found ->
     let filename = Getter.get uri in
      let (annobj,_) = CicParser.term_of_xml filename uri false in
       let obj = Deannotate.deannotate_obj annobj in
        let output = Unchecked obj in
         HashTable.add hashtable (uri,0) output ;
         output
;;

(* DANGEROUS!!!                                *)
(* USEFUL ONLY DURING THE FIXING OF THE FILES  *)
(* change_obj uri (Some newobj)                *)
(*  maps uri to newobj in cache.               *)
(* change_obj uri None                         *)
(*  maps uri to a freeze dummy-object.         *)
let change_obj uri newobj =
 let newobj =
  match newobj with
     Some newobj' -> Unchecked newobj'
   | None         -> Frozen (Cic.Variable ("frozen-dummy", None, Cic.Implicit))
 in
  HashTable.remove hashtable (uri,0) ;
  HashTable.add hashtable (uri,0) newobj
;;

let is_annotation_uri uri =
 Str.string_match (Str.regexp ".*\.ann$") (UriManager.string_of_uri uri) 0
;;

(* returns both the annotated and deannotated uncooked forms (plus the *)
(* map from ids to annotation targets)                                 *)
let get_annobj_and_type_checking_info uri =
 let filename = Getter.get uri in
  match CicParser.term_of_xml filename uri true with
     (_, None) -> raise Impossible
   | (annobj, Some ids_to_targets) ->
    (* If uri is the uri of an annotation, let's use the annotation file *)
    if is_annotation_uri uri  then
     AnnotationParser.annotate (Getter.get_ann uri) ids_to_targets ;
    try
      (annobj, ids_to_targets, HashTable.find hashtable (uri,0))
    with
     Not_found -> 
      let obj = Deannotate.deannotate_obj annobj in
       let output = Unchecked obj in
        HashTable.add hashtable (uri,0) output ;
        (annobj, ids_to_targets, output)
;;


(* get_obj uri                                                               *)
(* returns the cic object whose uri is uri. If the term is not just in cache, *)
(* then it is parsed via CicParser.term_of_xml from the file whose name is    *)
(* the result of Getter.get uri                                               *)
let get_obj uri =
 match get_obj_and_type_checking_info uri 0 with
    Unchecked obj -> obj
  | Frozen    obj -> obj
  | Cooked    obj -> obj
;;

(* get_annobj uri                                                             *)
(* returns the cic object whose uri is uri either in annotated and            *)
(* deannotated form. The term is put into the cache if it's not there yet.    *)
let get_annobj uri =
 let (ann, ids_to_targets, deann) = get_annobj_and_type_checking_info uri in
  let deannobj =
   match deann with
      Unchecked obj -> obj
    | Frozen    _   -> raise (CircularDependency (UriManager.string_of_uri uri))
    | Cooked    obj -> obj
  in
   (ann, ids_to_targets, deannobj)
;;

(*CSC Commento falso *)
(* get_obj uri                                                               *)
(* returns the cooked cic object whose uri is uri. The term must be present  *)
(* and cooked in cache                                                       *)
let rec get_cooked_obj uri cookingsno =
 match get_obj_and_type_checking_info uri cookingsno with
    Unchecked _
  | Frozen    _ -> raise UncookedObj
  | Cooked obj -> obj
;;

(* is_type_checked uri                                              *)
(* CSC: commento falso ed obsoleto *)
(* returns true if the term has been type-checked                   *)
(* otherwise it returns false and freeze the term for type-checking *)
(* set_type_checking_info must be called to unfreeze the term       *)
let is_type_checked uri cookingsno =
 match get_obj_and_type_checking_info uri cookingsno with
    Cooked obj -> CheckedObj obj
  | Unchecked obj ->
     HashTable.remove hashtable (uri,0) ;
     HashTable.add hashtable (uri,0) (Frozen obj) ;
     UncheckedObj obj
  | Frozen _ -> raise (CircularDependency (UriManager.string_of_uri uri))
;;

(* set_type_checking_info uri                               *)
(* must be called once the type-checking of uri is finished *)
(* The object whose uri is uri is unfreezed                 *)
let set_type_checking_info uri =
 match HashTable.find hashtable (uri,0) with
    Frozen obj ->
     (* let's cook the object at every level *)
     HashTable.remove hashtable (uri,0) ;
     let obj' = CicSubstitution.undebrujin_inductive_def uri obj in
      HashTable.add hashtable (uri,0) (Cooked obj') ;
      let cooked_objs = !cook_obj obj' uri in
       let last_cooked_level = ref 0 in
       let last_cooked_obj = ref obj' in
        List.iter
         (fun (n,cobj) ->
           for i = !last_cooked_level + 1 to n do
            HashTable.add hashtable (uri,i) (Cooked !last_cooked_obj)
           done ;
           HashTable.add hashtable (uri,n + 1) (Cooked cobj) ;
           last_cooked_level := n + 1 ;
           last_cooked_obj := cobj
         ) cooked_objs ;
        for i = !last_cooked_level + 1 to UriManager.depth_of_uri uri + 1 do
         HashTable.add hashtable (uri,i) (Cooked !last_cooked_obj)
        done
  | _ -> raise (CouldNotUnfreeze (UriManager.string_of_uri uri))
;;
